/*
 * Copyright (C) 2002 Luca Abeni
 * This is Free Software; see GPL.txt for details
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <muldiv.h>
#include <time.h>

#include <dispatch.h>
#include <cbs.h>
#include <log.h>
#include <trace.h>

static unsigned long long int prev_exec_set_time = 0;
static unsigned long long int last_update_time = 0;
static struct cbs_struct *exec = NULL;
static struct list_head ready_list;

static struct t_data_struct enforce;

#ifdef __GRUB__
static long int U = 0; /* Global bandwidth*/
#ifdef __GRUB_EXTRA_BANDWIDTH__
static long long int remaining = 0; /* Remaining bandwidth */
#define add_remaining(s) (s)->c += remaining; grub_dbg("%lld	%lld	%ld\n", (s)->c, remaining, U); remaining = 0
#else
#define add_remaining(s)
#endif
static inline long long int account(unsigned long long int time)
{
  long long int res;

  res = ((unsigned long int)llimd((time - last_update_time), U, fix_point));
  last_update_time = time;

  return res;
}
static inline void enforcement_set(struct cbs_struct *s)
{
  add_remaining(s);
  if (U <= 0) {
    sched_error("GRUB ERROR: U = 0!!!\n");
    U = 1;
  }
  if (U > fix_point) {
    sched_error("GRUB: U = %d!!! > 1\n", fix_point);\
  }
  grub_dbg("GRUB: enforce will fire in %lu(U=%ld)\n",
		  ((clock2us((unsigned long long int) ((s)->c)) ) * fix_point / U), U);
  sched_set_timer(llimd((s)->c, fix_point, U), &enforce);
}
#else
#define account(time) time - last_update_time; last_update_time = time
#define enforcement_set(s) sched_set_timer((s)->c, &enforce)
#endif

#ifdef __SRV_ID__
static int last_server_id = 1;
#endif

static int sched_caller = 0;

static inline void cbs_idle_cpu(int cpu)
{
#ifdef __MULTITASK__
  struct task_struct *t;
  struct list_head *h;
  int iter = 0;

  if (exec != NULL) {
    iter++;
    if (!list_empty(&exec->tasks)) {
      list_for_each(h, &exec->tasks) {
        t = (list_entry(h, struct task_list, others))->task;
        if (t->pid == 0) {
          sched_error("CBS ERROR!!! (idling) Idle is in queue? (iteration %d)\n", iter);
          sched_error("%p --- %p --- %d\n", list_entry(h, struct task_list, others),
		      (list_entry(h, struct task_list, others))->task,
		      t->pid);
        } else {
          stop(t);
        }
      }
    } 
  }
#else /* __MULTITASK__ */
  if (exec != NULL) {
    stop(exec->task);
  }
#endif /* __MULTITASK__ */
  
  gensched_idle_cpu(cpu);
}

static inline void cbs_dispatch(struct cbs_struct *c, unsigned long long int time, int p)
{
  struct task_struct *t;
#ifdef __MULTITASK__
  struct list_head *h;

  cbs_dbg("        descheduling task(s) ");
  if (exec != NULL) {
    if (!list_empty(&exec->tasks)) {
      list_for_each(h, &exec->tasks) {
        int iter = 0;

        iter++;
        t = (list_entry(h, struct task_list, others))->task;
        if (t->pid == 0) {
          sched_error("CBS ERROR!!! (descheduling) Idle is in queue? (iteration %d)\n", iter);
          sched_error("%p --- %p --- %d\n", list_entry(h, struct task_list, others),
		      (list_entry(h, struct task_list, others))->task,
		      t->pid);
        } else {
          cbs_dbg(" %d", t->pid);
	  stop(t);
        }
      }
    }
  }

  cbs_dbg("schedule() done: selected task(s) ");
  
  if (!list_empty(&c->tasks)) {
    list_for_each(h, &c->tasks) {
      int iter = 0;

      iter++;
      t = (list_entry(h, struct task_list, others))->task;
      if (t->pid == 0) {
        sched_error("CBS ERROR!!! Idle is in queue? (iteration %d)\n", iter);
        sched_error("%p --- %p --- %d\n", list_entry(h, struct task_list, others),
		      (list_entry(h, struct task_list, others))->task,
		      t->pid);
      } else {
        cbs_dbg(" %d", t->pid);

        dispatch(t, p);
      }
    }
  } else {
    printk("WTF??? Now tasks in CBS Queue!\n");
  }
  cbs_dbg("\n");
#else /* __MULTITASK__ */
  t = c->task;
  cbs_dbg(" %d\n", t->pid);
  if (exec != NULL) {
    stop(exec->task);
  }
  dispatch(t, p);
#endif /* __MULTITASK__ */

  trace_dispatch(c, time);
}

static int update_used_time(struct cbs_struct *cbs, unsigned long long int time)
{
  long long int consumed_time;

  cbs_dbg("        Update Used Time...\n");
  sched_stop_timer(&enforce);

  if (time < last_update_time) {
    sched_error("[%llu]ERROR: time going back!!! < %llu\n",
		    clock2us(time), clock2us(last_update_time));

    return 0;
  }
  consumed_time = account(time);
  if (consumed_time < 0) {
    sched_error("[%llu]ERROR: consumed time < 0\n", clock2us(time));
  }
  if (consumed_time > cbs->c) {
    sched_extra_warning("[%llu]WARNING: consumed time  = %lld > %lld\n",
		    clock2us(time), consumed_time, cbs->c);
  }
  cbs->c -= consumed_time;

  if (cbs->c <= 0) {
    sched_extra_warning("CBS WARNING: c < consumed... (%lu < %lu) (Done Anyway)\n",
		      (unsigned long int)cbs->c,
		      (unsigned long int)consumed_time);
    while (cbs->c <= 0) {
      cbs->c += cbs->max_budget_clock;
      cbs->deadline += cbs->period_clock;
      trace_dl_postpone(cbs, time,
		  cbs->deadline - cbs->period_clock, cbs->deadline);
    }
#ifdef __HARD_RSV__
    if (cbs->deadline - cbs->period_clock > time) {
      rsv_set_state(cbs, RSV_DEPLETED);
      if (cbs->rlist.next != NULL) {
        list_del(&(cbs->rlist)); /* Remove the server from the list */
        cbs->rlist.next = NULL;
        cbs->rlist.prev = NULL;
        sched_set_timer(cbs->deadline - cbs->period_clock - time, &(cbs->reactivate));
      
        return 0;
      }
    }
#endif

    return 1;
  }
  
  cbs_dbg("    Done the update\n");
  return 0;
}

/* Returns 1 if the task is inserted at the beginning of the ready queue,
 * Returns 0 otherwise
 */
static int edf_list_add(struct cbs_struct *t, int type)
{
  struct list_head *p;
  int done;

#if defined(__GRUB__) || defined(__HARD_RSV__)
  if (rsv_state(t) != RSV_ACTIVE) {
    sched_error("ERROR: Inserting in EDF Queue rsv with state %d\n",
            rsv_state(t));
    BUG();
  }
#endif

  cbs_dbg("        EDF List Add %d\n", type);
  if (ready_list.next == &ready_list) {
    list_add(&(t->rlist), &ready_list);
    
    return 1;
  } else {
    p = ready_list.next;
    done = 0;
    while (!done) {
      if (p == &ready_list) {
	done = 1;
      } else {
	if (list_entry(p, struct cbs_struct, rlist)->deadline > t->deadline) {
	  done = 1;
	} else {
          p = p->next;
	}
      }
    }
    list_add_tail(&(t->rlist), p);
  }

  cbs_dbg("    Done the edf add\n");
  if (&(t->rlist) == ready_list.next) {
    return 1;
  }

  return 0;
}

static int update_priorities(struct cbs_struct *cbs)
{
  if (cbs->rlist.next == NULL) {
    sched_error("CBS ERROR!!! Update priority on a task that is not in ready queue!!!\n");
    return 1;
  }
	      /* This can be null if we are invoked by deactivate() */
  list_del(&(cbs->rlist));
  return edf_list_add(cbs, 0);
}

static void cbs_schedule(unsigned long long int time)
{
  cbs_dbg("        CBS schedule()\n");

  if (exec == NULL) {
    if (current->private_data != NULL) {
      struct cbs_struct *c;

      c = current->private_data;
      if (c->new_task) {
	c->new_task = 0;
      } else {
        sched_extra_warning("[%d %ld] CBS WARNING: Exec = NULL but current->private_data != NULL! Caller = %d\n",
		      current->pid, current->rt_priority, sched_caller);
      }
    }
  } else {
    if ((current->private_data != NULL) && (current->private_data != exec)) {
      sched_extra_warning("[%d %ld] CBS WARNING: Exec != current->private_data (%p != %p)\n",
		      current->pid, current->rt_priority, exec, current->private_data);
#if 0
      return;
#endif /* Why this? */
    }
  }

  cbs_dbg("            Checking ready queue...\n");
  if (list_empty(&ready_list)) {
    cbs_dbg("                No CBS tasks in ready queue...\n");
    cbs_idle_cpu(0);
    exec = NULL;
    cbs_dbg("    schedule() done: no CBS task\n");
    trace_idle_dispatch(time);
  } else {
    struct cbs_struct *new_exec;
#ifdef __CBS_DEBUG__
    struct list_head *tmp;

    list_for_each(tmp, &ready_list) {
#ifdef __MULTITASK__
      cbs_dbg("%p: Dl = %Ld\n",
		    list_entry(tmp, struct cbs_struct, rlist),
		    list_entry(tmp, struct cbs_struct, rlist)->deadline);
#else
      cbs_dbg("%d: Dl = %Ld\n",
		    list_entry(tmp, struct cbs_struct, rlist)->task->pid,
		    list_entry(tmp, struct cbs_struct, rlist)->deadline);
#endif
    } 
  
    sched_print("                Getting first ready task...\n");
#endif

/*  t = list_entry(ready_list.next, struct cbs_struct, rlist)->task;*/
    new_exec = list_entry(ready_list.next, struct cbs_struct, rlist);
#ifdef __MULTITASK__
    cbs_dbg("            Selected CBS %p\n", new_exec);
#else
    cbs_dbg("            Selected CBS %p (task %d)\n", new_exec, new_exec->task->pid);
#endif
    cbs_dbg("            Setting timer in %Ld ticks\n", new_exec->c);

    enforcement_set(new_exec);
    
    cbs_dispatch(new_exec, time, 0);
    exec = new_exec;
#if defined(__GRUB__) || defined(__HARD_RSV__)
    if (rsv_state(exec) != RSV_ACTIVE) {
      sched_error("Exec state = %d (!= 1)!!!\n", rsv_state(exec));
      dump_stack();
    }
#endif
    prev_exec_set_time = time;
  } 
}

#ifdef __HARD_RSV__
/* This function starts when a Server rebecomes active after being blocked 
 * (when the "reactivate" timer fires) 
 * */
static void cbs_sleep_reactivate(unsigned long long int time, void* v)
{
  struct cbs_struct* t;

  if (v== NULL) {
    sched_error("CBS ERROR: parameter of cbs_sleep_reactivate is NULL\n");
  }
  cbs_dbg("Starting cbs_sleep_reactivate\n");
  t = (struct cbs_struct*) v;
  if (rsv_state(t) != RSV_DEPLETED) {
    sched_error("CBS ERROR: reactivating a non-depleted reserve?\n");
  }
  rsv_set_state(t, RSV_ACTIVE);
  if (edf_list_add(t, 0)) {
    /* We need a context switch */
    if (exec) {
      /* Stop the executing server */
      if (update_used_time(exec, time)) {
        update_priorities(exec);
      }
    } else {
      /* Noone was in execution... We need to explicitly set last_update_time,
       * because update_used_time() has not been called!
       */
      last_update_time = time;
    }
    sched_caller=6;
    cbs_schedule(time);
    sched_caller=0;
  }
}
#endif /* __HARD_RSV__ */

static void cbs_postpone(unsigned long long int time, void *dummy)
{
  struct cbs_struct *t;
  
  cbs_dbg("[%llu]Postpone: Current = %d\n", clock2us(time), current->pid);
  cbs_dbg("    Exec = %lx\n", (unsigned long int)exec);
  if (exec != NULL) {
#ifdef __MULTITASK__
    cbs_dbg("        (%p)\n", exec);
#else
    cbs_dbg("        (%d)\n", exec->task->pid);
#endif
  }
  if (exec == NULL) {
    sched_error("CBS ERROR!!! Postpone Deadline with Exec == NULL!!!\n");
  } else {
#ifndef __MULTITASK__
    if ((current != exec->task) && (current->rt_priority == 0)) {
      sched_warning("CBS WARNING: current (%d) != Exec (%d) in postpone()!!!\n",
		      current->pid, exec->task->pid);
    }
#endif
  }

  if (exec == NULL) {
    sched_print("I am completely screwed up... Returning :(\n");
    printk("I am completely screwed up... Returning :(\n");

    return;
  }
  t = exec;

  if (current->state != TASK_RUNNING) {
    sched_extra_warning("Uhmmm... Postpone on non-RUNNING task %d: %d... Returning!\n",
		    current->pid, current->state);
    return;
  }
#if defined(__GRUB__) || defined(__HARD_RSV__)
  if (rsv_state(t) != RSV_ACTIVE) {
    /* This can happen if:
     * 1) update_used_time() is called for some reason
     * 2) it finds that the budget is exhausted, and decides to deplete the rsv
     * 3) Before actually arriving to schedule(), the depletion timer fires...
     * ---> In this case is it safe to just return?
     */
    sched_error("CBS WARNING: Depleting a non-active reserve? %d\n", rsv_state(t));
    return;
  }
#endif

  if (update_used_time(t, time)) {
    if (update_priorities(t) == 0) {
      cbs_dbg("Postpone: will start another server\n");
      sched_caller = 1;
      cbs_schedule(time);
      sched_caller = 0;
    } else {
      enforcement_set(t);
    }
  } else {	    
    cbs_dbg("Postpone: I'm blocking...\n");
    sched_caller = 5;
    cbs_schedule(time);
    sched_caller = 0;
  }
  cbs_dbg("Done the Postpone\n");
}

#ifdef __GRUB__
/* This function starts when a Server becomes inactive */
static void grub_inactive(unsigned long long int time, void* v)
{   
  struct cbs_struct* t;
  int have_to_sched;

  if (v == NULL) {
    sched_error("GRUB ERROR: parameter of grub_inactive is NULL\n");

    return;
  }
  
  t = (struct cbs_struct*) v;
   
  if (rsv_state(t) == GRUB_NON_CONTENDING) {
    have_to_sched = 0;

    grub_dbg("GRUB: Inactivating task\n");
    if ((exec != NULL) && (exec!= t)) {
      grub_dbg("GRUB Inactive: updating used time\n");
      if (update_used_time(exec, time)) {
        if (update_priorities(exec) == 0) {
          have_to_sched = 1;
	}
      } else {	    
        grub_dbg("Blocking...\n");
        have_to_sched = 1;
      }
    }
    /* U-=Ui */
    grub_dbg("GRUB: U = %ld ->", U);
    U -= llimd(t->max_budget, fix_point, t->period);
    grub_dbg(" %ld\n", U);
    if (U < 0) {
      sched_error("GRUB ERROR(grub_inactive): U =%ld< 0\n",U);
      U = 0;
    }
    if (have_to_sched) {
      sched_caller = 1;	/* FIXME! */
      cbs_schedule(time);
      sched_caller = 0;
    } else if (exec != NULL) {
      enforcement_set(exec);
    }
    rsv_set_state(t, GRUB_INACTIVE); /* Server becomes inactive */
  } else {
    sched_error("GRUB ERROR: Grub Inactive on server with state = %d\n",
		    rsv_state(t));
  }
}
#endif


/*
 * Exported functions
 */

#if !defined __GRUB__
/* cbs_activate for CBS */
int cbs_activate(struct cbs_struct *t, unsigned long long r)
{
#ifdef __MULTITASK__
  cbs_dbg("[%llu]Activating CBS %lx\n", r, (unsigned long int)t);
  cbs_dbg("    Exec now is %p\n", exec);
  if (t->activations++ != 0) {
    cbs_dbg("Server is already active (%d)\n", t->activations);
    if (t == exec) {
      return 2;
    }
    return 0;
  }
#else
  cbs_dbg("Activating CBS %d\n", t->task->pid);
#endif

#ifdef __HARD_RSV__
  if (rsv_state(t) == STATE_NONE) {
    rsv_set_state(t, RSV_INACTIVE);
  }
  if (rsv_state(t) != RSV_INACTIVE) {
    sched_error("CBS ERROR: Activating rsv with state %d\n", rsv_state(t));
  }
#endif

  if (t->arr_time != 0) {
    sched_error("CBS ERROR: Arrival at %Lu with arrival rime already %Lu\n",
		    clock2us(r), clock2us(t->arr_time));
  }
  t->arr_time = r; 
  
  trace_activation(t, r);
  cbs_dbg("New DL?\n");
  /* CBS test: c > (d - r) * U ---> New deadline */
  if ((t->c > llimd(t->deadline - r, t->max_budget_clock, t->period_clock)) ||
		  (t->deadline < r)) {
    /* Generate new deadline */
    t->deadline = r + t->period_clock;
    t->c = t->max_budget_clock;
    cbs_dbg("    New budget %Ld\n",t->c);
    cbs_dbg("    New Deadline %Ld\n", t->deadline);
    trace_dl_assign(t, r, t->deadline);
  }
  rsv_set_state(t, RSV_ACTIVE);	/* Of for GRUB! RSV_ACTIVE == GRUB_CONTENDING */
  
  cbs_dbg("    Adding to list...");
  if (edf_list_add(t, 1)) {
    cbs_dbg("    Going to schedule...");
    if (exec) {
      if (update_used_time(exec, r)) {
        update_priorities(exec);
      }
    } else {
      /* Noone was in execution... We need to explicitly set last_update_time,
       * because update_used_time() has not been called!
       */
      last_update_time = r;
    }
    cbs_dbg("last_update_time = %llu (cbs_activate)\n",
		    clock2us(last_update_time));
    sched_caller = 2;
    cbs_schedule(r);
    sched_caller = 0;
  }
  cbs_dbg("Done the activate()\n");

  return (exec == t);
}
#else
/* cbs_activate for GRUB */
int cbs_activate(struct cbs_struct *t, unsigned long long r)
{
  int have_to_sched;

#ifdef __MULTITASK__
  grub_dbg("[%llu]Activating CBS %lx\n", clock2us(r), (unsigned long int)t);
  grub_dbg("    Exec now is %p\n", exec);
  if (t->activations++ != 0) {
    cbs_dbg("Server is already active (%d)\n", t->activations);
    if (t == exec) {
      return 2;
    }
    return 0;
  }
#else
  grub_dbg("Activating CBS %d\n", t->task->pid);
#endif

  if (rsv_state(t) == STATE_NONE) {
    rsv_set_state(t, RSV_INACTIVE);
  }

  if (t->arr_time != 0) {
    sched_error("CBS ERROR: Arrival at %Lu with arrival rime already %Lu\n",
		    clock2us(r), clock2us(t->arr_time));
  }
  t->arr_time = r; 
  
  trace_activation(t, r);
  have_to_sched = 0;
  if (exec != NULL) {
    grub_dbg("GRUB Inactive: updating used time\n");
    if (update_used_time(exec, r)) {
      update_priorities(exec);
      have_to_sched = 1;
    }
  } else {
    /* Noone was in execution... We need to explicitly set last_update_time,
     * because update_used_time() has not been called!
     */
    last_update_time = r;
    have_to_sched = 1;
  }
  grub_dbg("last_update_time = %llu (cbs_activate)\n",
		  clock2us(last_update_time));
  
  if (rsv_state(t) == GRUB_INACTIVE) {
    /* Server was Inactive */


    grub_dbg("GRUB: Activating an inactive Server\n");

    grub_dbg("GRUB: U = %ld ->", U);
    U += llimd(t->max_budget, fix_point, t->period);
    grub_dbg(" %ld\n", U);
    
    /* Generate new deadline */
    t->deadline = r + t->period_clock;
    t->c = t->max_budget_clock;

    if (U > fix_point) {
      sched_error("GRUB: U=%ld > 1\n",U);
      /* U = fix_point; */
    }
  } else if (rsv_state(t) == GRUB_NON_CONTENDING) {
    /* Server was ActiveNonContending */

    /* Since Vi = deadline-(c/bandwidth) we just postpone deadline of (period-(c/bandwidth)) */
    t->deadline += (unsigned long long int) ((t->period_clock) - llimd(t->c,t->period,t->max_budget));
    t->c = t->max_budget_clock;  
    sched_stop_timer(&(t->inactive));
    grub_dbg("GRUB: Stopping timer inactive\n");
    grub_dbg("GRUB: Activating an ActiveNonContending Server\n");
  } else {
    sched_error("GRUB ERROR: Reactivate with strange state %d\n", rsv_state(t));
  }
  rsv_set_state(t, GRUB_CONTENDING); /* Server becomes ActiveContending */
  grub_dbg("    New budget %ld\n",t->c);
  grub_dbg("    New Deadline %Ld\n", t->deadline);
  trace_dl_assign(t, r, t->deadline);

  if (edf_list_add(t, 1)) {
    have_to_sched = 1;
  }
  if (have_to_sched) {
    grub_dbg("    Going to schedule...");
    sched_caller = 2;
    cbs_schedule(r);
    sched_caller = 0;
  } else {
    enforcement_set(exec);
  }
  grub_dbg("Done the activate()\n");

  return (exec == t);
}
#endif

int cbs_deactivate(struct cbs_struct *t, unsigned long long int time)
{
  int next_state;
#ifdef __GRUB__
  long long int inactive_time;
  unsigned long long int virtual_time;
#endif

#ifdef __HARD_RSV__
  if (rsv_state(t) != RSV_DEPLETED) {
#else
  if (1) {
#endif
    if ((t->rlist.next==NULL) || (t->rlist.prev==NULL)) {
      sched_error("ERROR: cbs_deactivate() has t->rlist.next or t->rlist.prev equal to NULL\n");
#ifdef __MULTITASK__
      sched_error("Activations is %d\n", t->activations);
#endif
#ifdef __HARD_RSV__
      sched_error("State is %d\n", rsv_state(t));
#endif
      dump_stack();
      for(;;);
    }
  }

#ifdef __MULTITASK__
  cbs_dbg("[%llu]CBS deactivate(%p)\n", clock2us(time), t);
  cbs_dbg("    Exec now is %p\n", exec);
  if (t->activations < 1) {
    sched_error("    ERROR!!! CBS Deactivate with Activations = %d < 1\n",
		    t->activations);

    return 0;	/* Do not really know what to do, here... :( */
  }
  if (t->activations-- > 1) {
    cbs_dbg("    Server still active (%d)\n", t->activations);
    cbs_dbg("    Deadline = %llu\n", t->deadline);
    return 0;
  }
#else
  cbs_dbg("[%llu]CBS deactivate(%d)\n", clock2us(time), t->task->pid);
  cbs_dbg("    Exec now is %p\n", exec);
#endif

  next_state = RSV_INACTIVE;
  if (t->arr_time == 0) {
    sched_error("CBS ERROR: Deactivate %p with arrival time = 0\n", t);
  }
  t->arr_time = 0;
#ifdef __GRUB__
  virtual_time = t->deadline - llimd(t->c, t->period, t->max_budget);
  if (virtual_time > time) {
    /* Server becomes ActiveNonContending */
    next_state = GRUB_NON_CONTENDING; 
    /* The server will become inactive in (d-time)-(c/Ui) */
    inactive_time = t->deadline - time - llimd(t->c, t->period, t->max_budget);
    if (inactive_time <= 0) {
      inactive_time = 1;
    }
    sched_set_timer(inactive_time, &(t->inactive));
    grub_dbg("GRUB: Setting timer inactive\n");
    grub_dbg("GRUB: Deactivating task\n");
    grub_dbg("GRUB: timer inactive will fire in %lld\n", inactive_time);
  } else {
    /* Server becomes Inactive */
    //sched_print("Direct to inactive\n");

    /* U-=Ui */
    grub_dbg("GRUB: U = %ld ->", U);
    U -= llimd(t->max_budget, fix_point, t->period);
    grub_dbg(" %ld\n", U);
    if (U < 0) {
      sched_error("GRUB ERROR(cbs_deactivate): U =%ld< 0\n",U);
      U = 0;
    }
#ifdef __GRUB_EXTRA_BANDWIDTH__
    remaining = t->c; 
#endif
    /* Server becomes inactive, do not change next_state
     * (RSV_INACTIVE == GRUB_INACTIVE
     */
  }
#endif
  
  trace_deactivation(t, time);
  
#ifdef __HARD_RSV__
  if (rsv_state(t) != RSV_DEPLETED) {
#else
  if (1) {
#endif
    list_del(&(t->rlist));
    t->rlist.next = NULL;
    t->rlist.prev = NULL;
    if (exec) {
      update_used_time(exec, time);
    } else {
      last_update_time = time;
      sched_error("CBS ERROR!!! Deactivate with Exec == NULL???\n");
    }
    cbs_dbg("last_update_time = %llu (cbs_deactivate)\n",
		    clock2us(last_update_time));
  
    sched_caller = 3;
    cbs_schedule(time);
    sched_caller = 0;
  } else {
    /*sched_stop_timer(&t->reactivate);*/
    rsv_reactivation_stop(t);
  }
  rsv_set_state(t, next_state);
  cbs_dbg("Done the deactivate()\n");

  return 1;
}

struct cbs_struct *cbs_create(struct task_struct *task, int period, int max_budget)
{
  struct cbs_struct *p;
#ifdef __MULTITASK__
  struct task_list *task_list_entry;
#endif

  cbs_dbg("Creating CBS with period %d and maximum budget %d\n",
		  period, max_budget);
  
  p = kmalloc(sizeof(struct cbs_struct), GFP_ATOMIC);
  if (p == NULL) {
    return NULL;
  }

/*  p->rlist = kmalloc(sizeof(struct list_head), GFP_ATOMIC);
  if (p->rlist == NULL) {
    kfree(p);
    return NULL;
  }
  */
  p->rlist.next = NULL;
  p->rlist.prev = NULL;

  p->c = 0;
  p->deadline = 0;
  p->period = period;
  p->max_budget = max_budget;
  p->period_clock = us2clock(p->period);
  p->max_budget_clock = us2clock(p->max_budget);

#ifdef __GRUB__
  (p->inactive).callback = grub_inactive;
  (p->inactive).param=p;
  sched_init_timer(&(p->inactive));
  /* field state initialized with a not significant value (for debugging) */
#endif

#ifdef __HARD_RSV__
  (p->reactivate).callback = cbs_sleep_reactivate;
  (p->reactivate).param=p;
  sched_init_timer(&(p->reactivate));
#endif	/* __HARD_RSV__ */
  rsv_set_state(p, STATE_NONE);

  
  p->arr_time = 0;
#ifdef __MULTITASK__
  task_list_entry = kmalloc(sizeof(struct task_list), GFP_ATOMIC);
  if (task_list_entry == NULL) {
    kfree(p);
    return NULL;
  }
  INIT_LIST_HEAD(&p->tasks);
  list_add(&task_list_entry->others, &p->tasks);
  task_list_entry->task = task;
  p->activations = 0;
#ifdef __CBS_DEBUG__
  {
    struct list_head *h;
  
    h = p->tasks.next;
    cbs_dbg("    Adding %d to %p...", task->pid, p);
    cbs_dbg("      (%p = %p ?) ... ", h, &task_list_entry->others);
    cbs_dbg("      done: %d\n", (list_entry(h, struct task_list, others))->task->pid);
  }
#endif
#else
  p->task = task;
#endif
  p->new_task = 1;
  set_srv_id(p, last_server_id++);

  trace_creation(p, sched_read_clock());

  return p;
}

#ifdef __MULTITASK__
struct cbs_struct *cbs_add(struct task_struct *task, struct cbs_struct *cbs)
{
  struct task_list *task_list_entry;

  cbs_dbg("Adding task %d to CBS %p\n", task->pid, cbs);
  
  task_list_entry = kmalloc(sizeof(struct task_list), GFP_ATOMIC);
  if (task_list_entry == NULL) {
    return NULL;
  }
  list_add(&task_list_entry->others, &cbs->tasks);
  task_list_entry->task = task;
#ifdef __CBS_DEBUG__
  {
    struct list_head *h;
  
    h = cbs->tasks.next;
    cbs_dbg("    Adding %d to %p...", task->pid, cbs);
    cbs_dbg("     (%p = %p ?) ... ", h, &task_list_entry->others);
    cbs_dbg("     done: %d\n", (list_entry(h, struct task_list, others))->task->pid);
  }
#endif

  return cbs;
}

int cbs_remove(struct task_struct *task, struct cbs_struct *cbs)
{
  struct list_head *h, *h1;

  cbs_dbg("CBS remove %d from %p\n", task->pid, cbs);
  h1 = NULL;
  list_for_each(h, &cbs->tasks) {
    if ((list_entry(h, struct task_list, others))->task == task) {
      h1 = h;
    }
  }
  if (h1 == NULL) {
    sched_error("CBS ERROR!!! Removing task which is not in CBS?\n");
  } else {
    list_del(h1);
    kfree(list_entry(h1, struct task_list, others));
  }
  if (list_empty(&cbs->tasks)) {
    cbs_dbg("    List is empty\n");

    return 0;
  }
  cbs_dbg("    List is NOT empty\n");

  return 1;
}

#if !defined __GRUB__
/* cbs_reactivate for CBS */
int cbs_reactivate(struct cbs_struct *t, unsigned long long r)
{
  cbs_dbg("[%llu]ReActivating CBS %lx\n", r, (unsigned long int)t);
  cbs_dbg("    Exec now is %p\n", exec);
  if (t->activations++ != 0) {
    cbs_dbg("    Server is already active (%d)\n", t->activations);
    if (t == exec) {
      return 2;
    }
    return 0;
  }

  if (exec != NULL) {
    if (update_used_time(exec, r)) {
      update_priorities(exec);
    }
    cbs_idle_cpu(0);
    exec = NULL;
  } else {
#if 0
    /* Uhmmm... Why was this here???
     * FIXME: Figure it out!!!
     */
     printk("ERROR!!! Reactivating with exec = NUL!!!\n");
#endif
  }

#ifdef __HARD_RSV__
  if (rsv_state(t) == STATE_NONE) {
    rsv_set_state(t, RSV_INACTIVE);
  }
  if (rsv_state(t) != RSV_INACTIVE) {
    sched_error("CBS ERROR: ReActivating rsv with state %d\n", rsv_state(t));
  }
#endif

  if (t->arr_time != 0) {
    sched_error("CBS ERROR: Arrival at %Lu with arrival rime already %Lu\n",
		    clock2us(r), clock2us(t->arr_time));
  }
  t->arr_time = r; 
  
  trace_activation(t, r);
  /* CBS test: c > (d - r) * U ---> New deadline */
  if ((t->c > llimd(t->deadline - r, t->max_budget_clock, t->period_clock)) ||
		  (t->deadline < r)) {
	  
    /* Generate new deadline */
    t->deadline = r + t->period_clock;
    t->c = t->max_budget_clock;
    cbs_dbg("    New Deadline %Ld\n", t->deadline);
    trace_dl_assign(t, r, t->deadline);
  }

  rsv_set_state(t, RSV_ACTIVE);
  edf_list_add(t, 1);
  cbs_dbg("    Going to schedule...");
  sched_caller = 4;
  cbs_schedule(r);
  sched_caller = 0;
  cbs_dbg("Done the ReActivate()\n");

  return 1;
}

#else

/* cbs_reactivate for GRUB */
int cbs_reactivate(struct cbs_struct *t, unsigned long long r)
{
  grub_dbg("[%llu]ReActivating CBS %lx\n", clock2us(r), (unsigned long int)t);
  grub_dbg("    Exec now is %p\n", exec);
  if (t->activations++ != 0) {
    grub_dbg("    Server is already active (%d)\n", t->activations);
    if (t == exec) {
      return 2;
    }
    return 0;
  }

  if (exec != NULL) {
    if (update_used_time(exec, r)) {
      update_priorities(exec);
    }
    cbs_idle_cpu(0);
    exec = NULL;
  } else {
#if 0
    printk("ERROR!!! ReActivating with exec = NUL!!!\n");
#endif
  }

  t->arr_time = r; 
  trace_activation(t, r);

  if (rsv_state(t) == STATE_NONE) {
    rsv_set_state(t, GRUB_INACTIVE);
  }
  if (rsv_state(t) == GRUB_INACTIVE) {
    /* Server was Inactive */
    grub_dbg("GRUB: ReActivating an inactive Server\n");

    grub_dbg("GRUB: U = %ld ->", U);
    U += llimd(t->max_budget, fix_point, t->period);
    grub_dbg("%ld\n", U);
    /* Generate new deadline */
    t->deadline = r + t->period_clock;
    t->c = t->max_budget_clock;

    if (U > fix_point) {
      sched_error("GRUB: U=%ld > 1\n",U);
      /* U = fix_point; */
    }
  } else if (rsv_state(t) == GRUB_NON_CONTENDING) {
    /* Server was ActiveNonContending */

    /* Since Vi = deadline-(c/bandwidth) we just postpone deadline of (period-(c/bandwidth)) */
    t->deadline += (unsigned long long int) ((t->period_clock) - llimd(t->c,t->period,t->max_budget));
    t->c = t->max_budget_clock; 
    sched_stop_timer(&(t->inactive));
    grub_dbg("GRUB: Stopping timer inactive\n");
    grub_dbg("GRUB: ReActivating an ActiveNonContending Server\n");
  } else {
    sched_error("GRUB ERROR: Reactivate with strange state %d\n", rsv_state(t));
  }
  rsv_set_state(t, GRUB_CONTENDING); /* Server becomes ActiveContending */
  grub_dbg("    New budget %ld\n",t->c);
  grub_dbg("    New Deadline %Ld\n", t->deadline);

  edf_list_add(t, 1);
  grub_dbg("    Going to schedule...");
  sched_caller = 4;
  cbs_schedule(r);
  sched_caller = 0;
  grub_dbg("Done the ReActivate()\n");

  return 1;
}
#endif
#endif

void cbs_destroy(struct cbs_struct *t, unsigned long long int time)
{
#ifdef __GRUB__
  int have_to_sched;

  if (rsv_state(t) == GRUB_NON_CONTENDING || rsv_state(t) == GRUB_CONTENDING) {
    have_to_sched = 0;
    /* Server is still Active */

    if ((exec != NULL) && (exec!= t)) {

      grub_dbg("GRUB Inactive: updating used time\n");
      if (update_used_time(exec, time)) {
        if (update_priorities(exec) == 0) {
          have_to_sched = 1;
	}
      } else {	    
        grub_dbg("Blocking...\n");
        have_to_sched = 1;
      }
    }
    grub_dbg("GRUB: Destroying an ActiveNonContending Server\n");
    
    /* U-=Ui */
    grub_dbg("GRUB: U = %ld ->", U);
    U -= llimd(t->max_budget, fix_point, t->period);
    grub_dbg(" %ld\n", U);

    if (U < 0) {
      sched_error("GRUB ERROR(cbs_destroy): U=%ld < 0\n",U);
      U = 0;
    }

    if (have_to_sched) {
      sched_caller = 1;	/* FIXME! */
      cbs_schedule(time);
      sched_caller = 0;
    } else if (exec != NULL) {
      sched_error("GRUB ERROR in cbs_destroy(): have_to_sched = 0, but exec != NULL");
      enforcement_set(exec);
    }
    sched_stop_timer(&(t->inactive));  /* Stop the timer */
  } 
#endif

  cbs_dbg("CBS destroy(%p)\n", t);
  trace_destruction(t, sched_read_clock());
  if ((t->rlist.next != NULL) || (t->rlist.prev != NULL)) {
    sched_error("CBS ERROR!!! Destroying a task while in ready queue?\n");
    return;
  }

  kfree(t);
}

void cbs_init(void)
{
  INIT_LIST_HEAD(&ready_list);
  enforce.callback = cbs_postpone;
  enforce.param = NULL;
  sched_init_timer(&enforce);
  trace_start(sched_read_clock());
}
void cbs_deinit(void)
{
}
