/*
 * Copyright (C) 2002 Luca Abeni
 * This is Free Software; see GPL.txt for details
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include <cbs.h>
#include <scheduler.h>
#include <dispatch.h>
#include <muldiv.h>
#include <time.h>


EXPORT_SYMBOL_GPL(cbs_set_params);
EXPORT_SYMBOL_GPL(cbs_get_params);
int debug_code = 0;

spinlock_t generic_scheduler_lock __cacheline_aligned = SPIN_LOCK_UNLOCKED;

struct sched_param_new {
  struct sched_param sp;
  int size;
  void *p;
};

#ifdef __MULTITASK__
extern void cbsify(struct task_struct *t);
extern int linux_cbs(struct task_struct *t);
#endif

int cbs_get_params(struct task_struct *p, unsigned long int *qs, unsigned long int *ts)
{
  struct cbs_struct *cbs;
  unsigned long flags;
  
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  if (p == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  if (p->private_data == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  
  cbs = p->private_data;
  *ts = cbs->period;
  *qs = cbs->max_budget;
    
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}
int cbs_set_params(struct task_struct *p, unsigned long int qs, unsigned long int ts)
{
  struct cbs_struct *cbs;
  unsigned long flags;
  
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  if (p == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  if (p->private_data == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  
  cbs = p->private_data;
#ifdef QOSDBG
  cprintf("<changing period> Task (pid %d) changed from %d to %d\n",
		  p->pid, p->oldp, p->per);
  cprintf("    (quantum is %d)\n", cbs->max_budget);
#endif

  cbs->period = ts;
  cbs->period_clock = us2clock(cbs->period);

#ifdef QOSDBG
  cprintf("    changed to %d %d\n", cbs->max_budget, cbs->period);
#endif

#ifdef QOSDBG
  cprintf("      Task %d changed capacity from %d (%d) to %d\n",
		  p->pid, p->oldq, ce.max_budget, p->q);
#endif
  cbs->max_budget = qs;
  cbs->max_budget_clock = us2clock(cbs->max_budget);
  
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}

void generic_request(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;

#ifdef __CBS_DEBUG__
  sched_print("Task %d arrives\n", t->pid);
#endif
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 1;
  
  if (t->private_data != NULL) {
    int res;
    
    time = sched_read_clock();
    res = cbs_activate(t->private_data, time);
    if (res == 2) {
#ifdef __CBS_DEBUG__
      sched_print("    prio goes to %d\n", SCHEDULING_PRIORITY);
#endif
      t->rt_priority = SCHEDULING_PRIORITY;
    } else if (res == 0) {
      t->rt_priority = IDLE_PRIORITY;
    }
  }
  debug_code = 0; 
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
#ifdef __CBS_DEBUG__
  sched_print("Activation done...\n");
#endif
}

void generic_release(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;

#ifdef __CBS_DEBUG__
  sched_print("Task %d blocks\n", t->pid);
#endif
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 2; 
  if (t->private_data != NULL) {
    time = sched_read_clock();
    if (cbs_deactivate(t->private_data, time) == 0) {
      t->rt_priority = IDLE_PRIORITY;
    }
  }
  debug_code = 0; 
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
#ifdef __CBS_DEBUG__
  sched_print("Deactivation done...\n");
#endif
}

void generic_register(struct task_struct *t)
{
  unsigned long flags;

  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 3;
  if (t->private_data != NULL) {
#ifdef __EXTRA_WARNING__
    sched_print("Uhm... Task %d is borning with the wrong private_data %p...\n",
		    t->pid,
		    t->private_data);
#endif
    t->private_data = NULL;
  }
    
  /* Check if the task is borning wrong... */
  if (t->rt_priority == SCHEDULING_PRIORITY) {
#if 0
#ifdef __EXTRA_WARNING__
    sched_print("Uhm... Task %d is borning with the wrong rt_priority...\n",
		    t->pid);
#endif
    t->policy = SCHED_OTHER;
    t->rt_priority = 0;
#else
    sched_print("Erkkk!!! Task %d is borning with the wrong rt_priority... But I cannot do anything!", t->pid);
#endif
  }
#ifdef __MULTITASK__
  cbsify(t);
#if 0
  t->policy = SCHED_RR;
  t->rt_priority = IDLE_PRIORITY;
#else
  gensched_dispatch(t, GENSCHED_DISABLE);
#endif
  if (t->state == TASK_RUNNING) {
    generic_request(t);
  } 
#endif
  debug_code = 0;
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
}

void generic_unregister(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;

  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 4; 
  if (t->private_data != NULL) {
    time = sched_read_clock();
    /* t->state may be != TASK_RUNNING if the task is killed by sigkill... */
    if (t->state == TASK_RUNNING) {
      cbs_deactivate(t->private_data, time);
    }
#ifdef __MULTITASK__
    if (cbs_remove(t, t->private_data) == 0) {
#endif
      cbs_destroy(t->private_data);
#ifdef __MULTITASK__
    }
#endif

    t->rt_priority = 0;
    t->private_data = NULL;
#if 0
    t->policy = SCHED_OTHER;
#endif
  }
  debug_code = 0; 
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
}

/* Return value:
 * < 0 ---> Failure
 * = 0 ---> Success
 * > 0 ---> do the regular Linux thing...
 */
int generic_getsched(struct task_struct *p, struct sched_param *sp)
{
  struct cbs_error cbs_e;
  struct sched_param_new sched_p, *param;
  unsigned long flags;
  unsigned long long int epsilon;
  struct cbs_struct *cbs;
  struct arrival_list *arr;

  param = (struct sched_param_new *)sp;

  spin_lock_irqsave(&generic_scheduler_lock, flags);
  if (p == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -ESRCH;
  }
  if (p->private_data == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return 1;
  }

  if (copy_from_user(&sched_p, param, sizeof(struct sched_param_new)) < 0) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }
  if (sizeof(struct cbs_error) != sched_p.size) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return 1;
  }
  if (copy_from_user(&cbs_e, sched_p.p, sizeof(struct cbs_error)) < 0) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }
 
  if (cbs_e.signature != E_SIGNATURE) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return 1;
  }
  
  cbs = p->private_data;
#ifdef __CBS_GET_ERROR__
  arr = list_entry(cbs->act_times.next, struct arrival_list, others);
  epsilon = cbs->deadline - arr->r;
  cbs_e.e = clock2us(epsilon);
  cbs_e.max_budget = cbs->max_budget;
  cbs_e.period = cbs->period;
  if (copy_to_user(sched_p.p, &cbs_e, sizeof(struct cbs_error)) < 0) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }
#endif
  
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}

int generic_setsched(struct task_struct *p, int policy, struct sched_param *sp)
{
  unsigned long int t, q;
#ifdef __MULTITASK__
  int move = 0;
  struct cbs_struct *old_server = NULL;
#endif
  struct cbs_struct *cbs;
  struct cbs_param cbs_p;
  struct sched_param_new sched_p;
  unsigned long flags;
  struct sched_param_new *param;
  unsigned long long int time;

  param = (struct sched_param_new *)sp;

  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 5;
  if (p == NULL) {
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -ESRCH;
  }
  if (policy < 0) {
    if (p->private_data) {
      policy = SCHED_CBS;
    } else {
      policy = p->policy;
    }
  }

  if (policy != SCHED_CBS) {
    if (p->private_data != NULL) {
      /* FIXME: Check this!!! */
      unsigned long long int time;

      time = sched_read_clock();
      cbs_deactivate(p->private_data, time);
#ifdef __MULTITASK__
      if (cbs_remove(p, p->private_data) == 0) {
#endif
        cbs_destroy(p->private_data);
#ifdef __MULTITASK__
      }
#endif

#if 0
      p->policy = SCHED_OTHER;
#endif
      p->private_data = NULL;
    }
    debug_code = 0; 
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return 1;
  }

  if (copy_from_user(&sched_p, param, sizeof(struct sched_param_new)) < 0) {
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }
  if (sizeof(struct cbs_param) != sched_p.size) {
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }
  if (copy_from_user(&cbs_p, sched_p.p, sizeof(struct cbs_param)) < 0) {
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EFAULT;
  }

    /* Extra consinstency check... */
    if (cbs_p.signature != CBS_SIGNATURE) {
      debug_code = 0;
      spin_unlock_irqrestore(&generic_scheduler_lock, flags);
      return -EFAULT;
    }
  
  t = cbs_p.period;
  q = cbs_p.max_budget;

  if (p->private_data != NULL) {
#ifdef __MULTITASK__
    if (!linux_cbs(p)) {
#endif
       struct cbs_struct *cbs;

       cbs = p->private_data;
#ifdef __CBS_DEBUG__
       sched_print("%d is already CBS!!! Changing parameters %lu %lu to %lu %lu\n",
		       p->pid, cbs->max_budget, cbs->period, q, t);
#endif
       /* For the moment, let's change the parameters in the simplest way... */
       /* FIXME: Maybe we should use the deadctivate --> change --> activate sequence? */
       cbs->period = t;
       cbs->max_budget = q;
       cbs->period_clock = us2clock(cbs->period);
       cbs->max_budget_clock = us2clock(cbs->max_budget);

      debug_code = 0;
      spin_unlock_irqrestore(&generic_scheduler_lock, flags);
      return 0;
#ifdef __MULTITASK__
    }

    move = 1;
    old_server = p->private_data;
    p->private_data = NULL;
#endif
  }
  cbs = cbs_create(p, t, q);
  if (cbs == NULL) {
    sched_print("Cannot create CBS!!!\n");
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -EINVAL;
  }

  time = sched_read_clock();
#ifdef __MULTITASK__
  if (p->state == TASK_RUNNING) {
    cbs_deactivate(old_server, time);
  }
  cbs_remove(p, old_server);
#endif

#ifdef __CBS_DEBUG__
  sched_print("Transforming %d in CBS\n", p->pid);
#endif
  p->private_data = cbs;
#if 0
  p->policy = SCHED_FIFO;
  p->rt_priority = IDLE_PRIORITY;
#else
  gensched_dispatch(p, GENSCHED_ENABLE);
#endif
  if (p->state == TASK_RUNNING) {
#ifdef __MULTITASK__
    if (move) {
      cbs_reactivate(cbs, time);
    } else
#endif
      cbs_activate(cbs, time);
  }
  debug_code = 0;
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}
