/*
 * Copyright (C) 2002 Luca Abeni
 * This is Free Software; see GPL.txt for details
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include <cbs-interface.h>
#include <cbs.h>
#include <dispatch.h>
#include <log.h>
#include <muldiv.h>
#include <time.h>


EXPORT_SYMBOL_GPL(cbs_set_params);
EXPORT_SYMBOL_GPL(cbs_get_params);
int debug_code = 0;

spinlock_t generic_scheduler_lock __cacheline_aligned = SPIN_LOCK_UNLOCKED;

struct sched_param_new {
  struct sched_param sp;
  int size;
  void *p;
};

#ifdef __MULTITASK__
extern void cbsify(struct task_struct *t);
extern int linux_cbs(struct task_struct *t);
#endif

int cbs_get_params(struct task_struct *p, unsigned long int *qs, unsigned long int *ts)
{
  struct cbs_struct *cbs;
  unsigned long flags;
  
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  if (p == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  if (p->private_data == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  
  cbs = p->private_data;
  *ts = cbs->period;
  *qs = cbs->max_budget;
    
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}
int cbs_set_params(struct task_struct *p, unsigned long int qs, unsigned long int ts)
{
  struct cbs_struct *cbs;
  unsigned long flags;
  
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  if (p == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  if (p->private_data == NULL) {
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    return -1;
  }
  
  cbs = p->private_data;
  cbs_dbg("<changing period> Task %d (pid ld) changed from %ld to %ld\n",
		  p->pid, cbs->period, ts);
  cbs_dbg("    (quantum is %ld)\n", cbs->max_budget);

  cbs->period = ts;
  cbs->period_clock = us2clock(cbs->period);

  cbs_dbg("    changed to %ld %ld\n", cbs->max_budget, cbs->period);

  cbs_dbg("      Task %d changed capacity from %ld to %ld\n",
		  p->pid, cbs->max_budget, qs);
  cbs->max_budget = qs;
  cbs->max_budget_clock = us2clock(cbs->max_budget);
  
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}

void generic_request(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;

  if (t->state == 0) {
    sched_print("[%d]Request: S=%ld\n", t->pid, t->state);
    dump_stack();
    for(;;);
  }
  t->state = TASK_RUNNING;
  cbs_dbg("Task %d arrives\n", t->pid);
  time = sched_read_clock();
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 1;
  
  if (t->private_data != NULL) {
    int res;
    
    res = cbs_activate(t->private_data, time);
    if (res == 2) {
      cbs_dbg("    prio goes to %d\n", CBS_DISPATCH_PRIORITY);
      dispatch(t, 0);
    } else if (res == 0) {
      stop(t);
    }
  }
  debug_code = 0; 
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
  cbs_dbg("Activation done...\n");
}

void generic_release(struct task_struct *t)
{
  unsigned long long int time;

  if (t->state & GENSCHED_DISABLED) {
    return;
  }
  cbs_dbg("Task %d blocks\n", t->pid);
  if (!irqs_disabled()) {
    /* This should _NEVER_ happen... We panic! */
    printk("CBS ERROR: Entering in generic_release with interrupts enabled!\n");
    dump_stack();
    for(;;);
  }
  time = sched_read_clock();
  
  /* Interrupts should be already disabled... But let's do the _irq thing,
   * because sched_read_clock() might have reenabled them! :(
   */
  spin_lock_irq(&generic_scheduler_lock);
  if (t->state == TASK_RUNNING) {
    printk("Something happened before we got the lock!\n");
    spin_unlock(&generic_scheduler_lock);
    return;
  }
  debug_code = 2; 
  if (t->private_data != NULL) {
    if (cbs_deactivate(t->private_data, time) == 0) {
      stop(t);
    }
  }
  debug_code = 0; 
  spin_unlock(&generic_scheduler_lock);
  cbs_dbg("Deactivation done...\n");
}

void generic_register(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;
#ifdef __MULTITASK__
  long int task_state;
#endif

  time = sched_read_clock();
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 3;
  if (t->private_data != NULL) {
    sched_extra_warning("Uhm... Task %d is borning with the wrong private_data %p...\n",
		    t->pid,
		    t->private_data);
    t->private_data = NULL;
  }
    
  /* Check if the task is borning wrong... */
  if (t->rt_priority == CBS_DISPATCH_PRIORITY) {
    sched_extra_warning("WARNING: Task %d is borning with priority %d (this can be ok)\n",
		    t->pid, t->rt_priority);
#ifndef __MULTITASK__
    genschedify(t, -1);
  }
#else
  } else {
    genschedify(t, CBS_DISPATCH_PRIORITY);
  }
  cbsify(t);
  task_state = t->state;
  stop(t);
  if (task_state == TASK_RUNNING) {
    /* generic_request(t); */
    int res;
    
    /* t->state = TASK_RUNNING;		FIXME: Is this needed? */
    res = cbs_activate(t->private_data, time);
    if (res == 2) {
      dispatch(t, 0);
    } else if (res == 0) {
      stop(t);
    }
  } 
#endif
  debug_code = 0;
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
}

void generic_unregister(struct task_struct *t)
{
  unsigned long flags;
  unsigned long long int time;
  unsigned state;

  time = sched_read_clock();
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 4; 
  if (t->private_data != NULL) {
    state = t->state;
    /* t->state may be != TASK_RUNNING if the task is killed by sigkill... */
    if ((t->state == TASK_RUNNING) || (t->state == GENSCHED_DISABLED)) {
      cbs_deactivate(t->private_data, time);
      state = TASK_RUNNING;
    }
#ifdef __MULTITASK__
    if (cbs_remove(t, t->private_data) == 0) {
#endif
      cbs_destroy(t->private_data, time);
#ifdef __MULTITASK__
    }
#endif

    t->private_data = NULL;
    t->state = state;
    genschedify(t, -1);
  }
  debug_code = 0; 
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
}

/* Return value:
 * < 0 ---> Failure
 * = 0 ---> Success
 * > 0 ---> do the regular Linux thing...
 */
int generic_getsched(struct task_struct *p, struct sched_param *sp)
{
  struct cbs_error cbs_e;
  struct sched_param_new sched_p, *param;
  unsigned long flags;
  unsigned long long int epsilon;
  unsigned long long int dl, r;
  unsigned long int q, t;
  struct cbs_struct *cbs;

  if (p->private_data == NULL) {
    return 1;
  }
  param = (struct sched_param_new *)sp;
  if (copy_from_user(&sched_p, param, sizeof(struct sched_param_new)) < 0) {
    return -EFAULT;
  }
  if (sizeof(struct cbs_error) != sched_p.size) {
    return 1;
  }
  if (copy_from_user(&cbs_e, sched_p.p, sizeof(struct cbs_error)) < 0) {
    return -EFAULT;
  }
 
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  cbs = p->private_data;
  dl = cbs->deadline;
  r = cbs->arr_time;
  q = cbs->max_budget;
  t = cbs->period;
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);
  if (cbs_e.signature == E_SIGNATURE) {
    epsilon = dl - r;
    cbs_e.e = clock2us(epsilon);
    cbs_e.max_budget = q;
    cbs_e.period = t;
    if (copy_to_user(sched_p.p, &cbs_e, sizeof(struct cbs_error)) < 0) {
      return -EFAULT;
    }

    return 0;
  }

  if (cbs_e.signature == E1_SIGNATURE) {
    epsilon = dl;
    cbs_e.e = clock2us(epsilon);
    cbs_e.max_budget = q;
    cbs_e.period = t;
    cbs_e.t = sched_read_clock();
    if (copy_to_user(sched_p.p, &cbs_e, sizeof(struct cbs_error)) < 0) {
      return -EFAULT;
    }
  
    return 0;
  }
  
  return 1;
}

int generic_setsched(struct task_struct *p, int policy, struct sched_param *sp)
{
  unsigned long int t, q;
#ifdef __MULTITASK__
  int move = 0;
  struct cbs_struct *old_server = NULL;
#endif
  struct cbs_struct *cbs;
  struct cbs_param cbs_p;
  struct sched_param_new sched_p;
  unsigned long flags;
  struct sched_param_new *param;
  unsigned long long int time;
  long int task_state;

  param = (struct sched_param_new *)sp;

  if (policy < 0) {
    if (p->private_data) {
      policy = SCHED_CBS;
    } else {
      policy = p->policy;
    }
  }

  if (policy != SCHED_CBS) {
    if (p->private_data != NULL) {
      /* FIXME: Check this!!! */
      unsigned long long int time;

      time = sched_read_clock();
      spin_lock_irqsave(&generic_scheduler_lock, flags);
      debug_code = 5;
      cbs_deactivate(p->private_data, time);
#ifdef __MULTITASK__
      if (cbs_remove(p, p->private_data) == 0) {
#endif
        cbs_destroy(p->private_data, time);
#ifdef __MULTITASK__
      }
#endif

      p->private_data = NULL;
      debug_code = 0; 
      spin_unlock_irqrestore(&generic_scheduler_lock, flags);
    }
    
    return 1;
  }

  if (copy_from_user(&sched_p, param, sizeof(struct sched_param_new)) < 0) {
    
    return -EFAULT;
  }
  if (sizeof(struct cbs_param) != sched_p.size) {
    
    return -EFAULT;
  }
  if (copy_from_user(&cbs_p, sched_p.p, sizeof(struct cbs_param)) < 0) {
    
    return -EFAULT;
  }

  /* Extra consinstency check... */
  if (cbs_p.signature != CBS_SIGNATURE) {
    
    return -EFAULT;
  }
  
  t = cbs_p.period;
  q = cbs_p.max_budget;
  if ((q < 1000) || (q > t)) {
    sched_error("ERROR: Creating a (%lu, %lu) reservation\n", q, t);

    return -EINVAL;
  }
  
  time = sched_read_clock();
  spin_lock_irqsave(&generic_scheduler_lock, flags);
  debug_code = 5;
  
  task_state = p->state;
  if (p->private_data != NULL) {
#ifdef __MULTITASK__
    if (!linux_cbs(p)) {
#endif
       struct cbs_struct *cbs;

       cbs = p->private_data;
       cbs_dbg("%d is already CBS!!! Changing parameters %lu %lu to %lu %lu\n",
		       p->pid, cbs->max_budget, cbs->period, q, t);
       /* For the moment, let's change the parameters in the simplest way... */
       /* FIXME: Maybe we should use the deadctivate --> change --> activate sequence? */
       cbs->period = t;
       cbs->max_budget = q;
       cbs->period_clock = us2clock(cbs->period);
       cbs->max_budget_clock = us2clock(cbs->max_budget);

      debug_code = 0;
      spin_unlock_irqrestore(&generic_scheduler_lock, flags);
      return 0;
#ifdef __MULTITASK__
    }

    move = 1;
    old_server = p->private_data;
    p->private_data = NULL;
#endif
  }
  cbs = cbs_create(p, t, q);
  if (cbs == NULL) {
    sched_print("Cannot create CBS!!!\n");
    debug_code = 0;
    spin_unlock_irqrestore(&generic_scheduler_lock, flags);

    return -EINVAL;
  }

#ifdef __MULTITASK__
  if (task_state == TASK_RUNNING) {
    cbs_deactivate(old_server, time);
  }
  cbs_remove(p, old_server);
#endif

  cbs_dbg("Transforming %d in CBS\n", p->pid);
  genschedify(p, CBS_DISPATCH_PRIORITY);
  p->private_data = cbs;
  stop(p);
  if (task_state == TASK_RUNNING) {
#ifdef __MULTITASK__
    if (move) {
      cbs_reactivate(cbs, time);
    } else
#endif
      cbs_activate(cbs, time);
  }
  debug_code = 0;
  spin_unlock_irqrestore(&generic_scheduler_lock, flags);

  return 0;
}
