/*
 * Copyright (C) 2002 Luca Abeni
 * This is Free Software; see GPL.txt for details
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#include <cbs.h>
#include <scheduler.h>

#ifndef CONFIG_GENERIC_SCHEDULER
#error The Generic Scheduler is not enabled!!!
#endif

MODULE_LICENSE("GPL");

extern void generic_request(struct task_struct *t);
extern void generic_release(struct task_struct *t);
extern void generic_register(struct task_struct *t);
extern void generic_unregister(struct task_struct *t);
extern int generic_setsched(struct task_struct *t, int policy, struct sched_param *param);
extern int generic_getsched(struct task_struct *t, struct sched_param *param);


#ifdef __MULTITASK__
static struct cbs_struct *lin_cbs;
#define LIN_Q 9000
#define LIN_T 90000
static int lin_qs, lin_ts;
MODULE_PARM(lin_qs, "i");
MODULE_PARM_DESC(lin_qs, "Maximum budget for the default CBS");
MODULE_PARM(lin_ts, "i");
MODULE_PARM_DESC(lin_ts, "Default CBS period");

static void create_linux_cbs(struct task_struct *t)
{
  if ((lin_qs == 0) || (lin_ts == 0)) {
    printk("Creating Linux CBS with default parameters (%d, %d)\n",
		    LIN_Q, LIN_T);
    lin_cbs = cbs_create(t, LIN_T, LIN_Q);
  } else {
    printk("Creating Linux CBS with default parameters (%d, %d)\n",
		    lin_qs, lin_ts);
    lin_cbs = cbs_create(t, lin_ts, lin_qs);
  }
  t->private_data = lin_cbs;
}

void cbsify(struct task_struct *t)
{
#ifdef __CBS_DEBUG__
  printk("Adding to Linux CBS\n");
#endif
  cbs_add(t, lin_cbs);
  t->private_data = lin_cbs;
}

int linux_cbs(struct task_struct *t)
{
  return (t->private_data == lin_cbs);
}
#endif

int init_module(void)
{
  struct task_struct *t, *dummy;

  if (block_hook != NULL) {
    printk("Error installing generic scheduler: block_hook != NULL!!!\n");
    return -1;
  }
  if (unblock_hook != NULL) {
    printk("Error installing generic scheduler: unblock_hook != NULL!!!\n");
    return -1;
  }
  if (fork_hook != NULL) {
    printk("Error installing generic scheduler: fork_hook != NULL!!!\n");
    return -1;
  }
  if (cleanup_hook != NULL) {
    printk("Error installing generic scheduler: cleanup_hook != NULL!!!\n");
    return -1;
  }
  if (setsched_hook != NULL) {
    printk("Error installing generic scheduler: setsched_hook != NULL!!!\n");
    return -1;
  }
  if (getsched_hook != NULL) {
    printk("Error installing generic scheduler: getsched_hook != NULL!!!\n");
    return -1;
  }

  block_hook = generic_release;
  unblock_hook = generic_request;
  fork_hook = generic_register;
  cleanup_hook = generic_unregister;
  setsched_hook = generic_setsched;
  getsched_hook = generic_getsched;

  cbs_init();
  /* If all the tasks must be managed by the generic scheduler, we must
   * ``transform'' all the existing tasks in CBS tasks...
   * We'll do it later.
   */
 
  do_each_thread(dummy, t) {
#ifdef __FP_DEAMONS__
    if (t->tty == NULL) {
#ifdef __SCHED_DEBUG__
      sched_print("Giving T%d the highest priority\n", t->pid);
#endif
      printk("Giving T%d the highest priority\n", t->pid);
      t->policy = SCHED_FIFO;
      t->rt_priority = SCHEDULING_PRIORITY + 1;
      t->private_data = NULL;
    } else
#endif
    {
#ifdef __MULTITASK__
      if(lin_cbs == NULL) {
	create_linux_cbs(t);
      } else {
        cbsify(t);
      }
      t->policy = SCHED_RR;
      t->rt_priority = IDLE_PRIORITY;

      if (t->state == TASK_RUNNING) {
	generic_request(t);
	/*
	unsigned long long int time;
	
	time = sched_read_clock();
	cbs_activate(lin_cbs, time);
	*/
      }
#endif
    }
  } while_each_thread(dummy, t);

  return 0;
}

void cleanup_module(void)
{
  struct task_struct *t, *dummy;

  t = current;
  
  do_each_thread(dummy, t) {
    printk("Deinitializing task %d    ", t->pid);
    if (t->private_data != NULL) {
      printk("Handled by generic scheduler!");
#if 1
      if (t->state == TASK_RUNNING) {
	generic_release(t);
      }

#ifdef __MULTITASK__
      if (cbs_remove(t, t->private_data) == 0)
#endif
        cbs_destroy(t->private_data);

      t->private_data = NULL;
#if 0
      t->policy = SCHED_OTHER;
      t->rt_priority = 0;
#endif
#else
      generic_unregister(t); /* We cannot, because the task may be inactive... */
#endif
    }
    printk("\n");
  } while_each_thread(dummy, t);
  cbs_deinit();
    
  block_hook = NULL;
  unblock_hook = NULL;
  fork_hook = NULL;
  cleanup_hook = NULL;
  setsched_hook = NULL;
  getsched_hook = NULL;
}
