/*
 * Copyright (C) 2002 Luca Abeni
 * This is Free Software; see GPL.txt for details
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#include <cbs.h>
#include <dispatch.h>
#include <log.h>

#include <muldiv.h>
#include <time.h>	/* For sched_read_clock()... Can we avoid it? */

#define __FP_DEAMONS__
#ifndef CONFIG_GENERIC_SCHEDULER
#error The Generic Scheduler is not enabled!!!
#endif

MODULE_LICENSE("GPL");

extern void generic_request(struct task_struct *t);
extern void generic_release(struct task_struct *t);
extern void generic_register(struct task_struct *t);
extern void generic_unregister(struct task_struct *t);
extern int generic_setsched(struct task_struct *p, int policy, struct sched_param *sp);
extern int generic_getsched(struct task_struct *p, struct sched_param *sp);


#ifdef __HARD_RSV__
static char *hard_parm = "HARD ";
#else
static char *hard_parm = "";
#endif

#ifdef __GRUB__
static char *grub_parm = "GRUB ";
#else
static char *grub_parm = "";
#endif

#ifdef __MULTITASK__
static char *multi_parm = "MULTI ";
static struct cbs_struct *lin_cbs;
#define LIN_Q 50000
#define LIN_T 500000
static int lin_qs, lin_ts;
module_param(lin_qs, int, 0);	/* perm = 0 ---> No sysfs */
module_param(lin_ts, int, 0);	/* perm = 0 ---> No sysfs */
MODULE_PARM_DESC(lin_ts, "Default CBS period");
MODULE_PARM_DESC(lin_qs, "Maximum budget for the default CBS");

static void create_linux_cbs(struct task_struct *t)
{
  if ((lin_qs == 0) || (lin_ts == 0)) {
    printk("Creating Linux CBS with default parameters (%d, %d)\n",
		    LIN_Q, LIN_T);
    lin_cbs = cbs_create(t, LIN_T, LIN_Q);
  } else {
    printk("Creating Linux CBS with default parameters (%d, %d)\n",
		    lin_qs, lin_ts);
    lin_cbs = cbs_create(t, lin_ts, lin_qs);
  }
  t->private_data = lin_cbs;
}

void cbsify(struct task_struct *t)
{
  cbs_dbg("Adding to Linux CBS\n");
  cbs_add(t, lin_cbs);
  t->private_data = lin_cbs;
}

int linux_cbs(struct task_struct *t)
{
  return (t->private_data == lin_cbs);
}
#else
static char *multi_parm = "";
#endif

static int cbsched_init(void)
{
  struct task_struct *t, *dummy;

  printk("Inserting CBS Module (parameters: [%s%s%s])\n",
		  multi_parm, grub_parm, hard_parm);
  if (block_hook != NULL) {
    printk("Error installing generic scheduler: block_hook != NULL!!!\n");
    return -1;
  }
  if (unblock_hook != NULL) {
    printk("Error installing generic scheduler: unblock_hook != NULL!!!\n");
    return -1;
  }
  if (fork_hook != NULL) {
    printk("Error installing generic scheduler: fork_hook != NULL!!!\n");
    return -1;
  }
  if (cleanup_hook != NULL) {
    printk("Error installing generic scheduler: cleanup_hook != NULL!!!\n");
    return -1;
  }
  if (setsched_hook != NULL) {
    printk("Error installing generic scheduler: setsched_hook != NULL!!!\n");
    return -1;
  }
  if (getsched_hook != NULL) {
    printk("Error installing generic scheduler: getsched_hook != NULL!!!\n");
    return -1;
  }

  block_hook = generic_release;
  unblock_hook = generic_request;
  fork_hook = generic_register;
  cleanup_hook = generic_unregister;
  setsched_hook = generic_setsched;
  getsched_hook = generic_getsched;

  cbs_init();
  /* If all the tasks must be managed by the generic scheduler, we must
   * ``transform'' all the existing tasks in CBS tasks...
   * We'll do it later.
   */
 
  do_each_thread(dummy, t) {
#ifdef __FP_DEAMONS__
    if (t->policy != SCHED_NORMAL) {
      printk("T%d is already RT(%lu): leaving it alone\n", t->pid, t->policy);
    } else
#endif
    {
#ifdef __MULTITASK__
      int is_active;

      genschedify(t, CBS_DISPATCH_PRIORITY);
      if(lin_cbs == NULL) {
	create_linux_cbs(t);
      } else {
        cbsify(t);
      }

      is_active = (t->state == TASK_RUNNING);
      stop(t);
      if (is_active) {
	generic_request(t);
	/*
	unsigned long long int time;
	
	time = sched_read_clock();
	cbs_activate(lin_cbs, time);
	*/
      }
#endif
    }
  } while_each_thread(dummy, t);

  return 0;
}

static void cbsched_cleanup(void)
{
  struct task_struct *t, *dummy;
  unsigned long long int time;

  printk("Removing CBS module... ");
  t = current;
  time = sched_read_clock();

  local_irq_disable();

  fork_hook = NULL;
  setsched_hook = NULL;
  getsched_hook = NULL;
  cleanup_hook = NULL;
  unblock_hook = NULL;
  block_hook = NULL;

  do_each_thread(dummy, t) {
    cbs_dbg("Deinitializing task %d    ", t->pid);
    if (t->private_data != NULL) {
      cbs_dbg("Handled by generic scheduler!");
      /* We cannot use gensched_unregister(), because the task may be inactive... */
      if (t->state == TASK_RUNNING) {
	while (t->state == TASK_RUNNING) {
	  t->state = TASK_INTERRUPTIBLE;
	  generic_release(t);
	}
	t->state = TASK_RUNNING;  /*Generic Release can change the task state...*/
      }

#ifdef __MULTITASK__
      if (cbs_remove(t, t->private_data) == 0)
#endif
        cbs_destroy(t->private_data, time);

      t->private_data = NULL;
      genschedify(t, -1);
    }
    cbs_dbg("\n");
  } while_each_thread(dummy, t);
  local_irq_enable();
  cbs_deinit();
 
  printk("Done\n");
}

module_init(cbsched_init);
module_exit(cbsched_cleanup);
